"""Test helper functions."""
from adaptavist.const import STATUS_BLOCKED, STATUS_FAIL, STATUS_IN_PROGRESS, STATUS_NOT_EXECUTED, STATUS_PASS
from bs4 import BeautifulSoup

from pytest_adaptavist._helpers import calc_test_result_status, html_row


class TestHelpersUnit:
    """Test helper functions on unit test level."""

    def test_calc_test_result_status(self):
        """Test calculation of test results."""
        assert calc_test_result_status([]) == STATUS_NOT_EXECUTED

        assert calc_test_result_status([{"status": STATUS_BLOCKED}, {"status": STATUS_NOT_EXECUTED}]) == STATUS_BLOCKED
        assert calc_test_result_status([{"status": STATUS_BLOCKED}, {"status": STATUS_IN_PROGRESS}]) == STATUS_BLOCKED
        assert calc_test_result_status([{"status": STATUS_BLOCKED}, {"status": STATUS_PASS}]) == STATUS_BLOCKED
        assert calc_test_result_status([{"status": STATUS_BLOCKED}, {"status": STATUS_FAIL}]) == STATUS_FAIL

        assert calc_test_result_status([{"status": STATUS_FAIL}, {"status": STATUS_NOT_EXECUTED}]) == STATUS_FAIL
        assert calc_test_result_status([{"status": STATUS_FAIL}, {"status": STATUS_IN_PROGRESS}]) == STATUS_FAIL
        assert calc_test_result_status([{"status": STATUS_FAIL}, {"status": STATUS_PASS}]) == STATUS_FAIL

        assert calc_test_result_status([{"status": STATUS_PASS}, {"status": STATUS_NOT_EXECUTED}]) == STATUS_IN_PROGRESS
        assert calc_test_result_status([{"status": STATUS_PASS}, {"status": STATUS_IN_PROGRESS}]) == STATUS_IN_PROGRESS

        assert calc_test_result_status([{"status": STATUS_IN_PROGRESS}, {"status": STATUS_NOT_EXECUTED}]) == STATUS_IN_PROGRESS

    def test_html_row(self):
        """Test html status row to be displayed in test case results."""
        assert html_row("passed", "") == ""
        assert html_row("failed", "") == ""
        assert html_row("blocked", "") == ""

        span = BeautifulSoup(html_row("passed", "Testmessage"), features="lxml").find("span")
        assert span.text == STATUS_PASS
        assert span.next_sibling == "Testmessage"
        assert "rgb(58, 187, 75)" in span.attrs["style"]

        span = BeautifulSoup(html_row("failed", "Testmessage"), features="lxml").find("span")
        assert span.text == STATUS_FAIL
        assert span.next_sibling == "Testmessage"
        assert "rgb(223, 47, 54)" in span.attrs["style"]

        span = BeautifulSoup(html_row("blocked", "Testmessage"), features="lxml").find("span")
        assert span.text == STATUS_BLOCKED
        assert span.next_sibling == "Testmessage"
        assert "rgb(75, 136, 231)" in span.attrs["style"]



"""Unit and system tests."""

from __future__ import annotations

import json
from typing import Any, Tuple
from unittest.mock import MagicMock

import requests
from _pytest.pytester import HookRecorder

from pytest_adaptavist._atm_configuration import ATMConfiguration

AdaptavistMock = Tuple[MagicMock, MagicMock, MagicMock]


def system_test_preconditions() -> bool:
    """Check preconditions for system tests."""
    atmcfg = ATMConfiguration()
    if (
        not atmcfg.global_config
        or "project_key" not in atmcfg.global_config
        or "jira_server" not in atmcfg.global_config
        or "jira_username" not in atmcfg.global_config
        or "jira_password" not in atmcfg.global_config
    ):
        return False

    try:
        requests.adapters.DEFAULT_RETRIES = 1
        requests.get(atmcfg.global_config.get("jira_server"), timeout=1)  # type: ignore
        return True
    except Exception:  # pylint: disable=broad-except
        return False


def get_test_values(report: HookRecorder, test_case: str = "") -> tuple[str, str]:
    """Get test run key and test case name."""
    user_properties = dict(report.matchreport(test_case).user_properties)
    test_run_key = user_properties["atmcfg"]["test_run_key"]
    test_name = user_properties["report"]["test_case_key"]
    return test_run_key, test_name


def read_global_config() -> dict[str, Any]:
    """Read global config and return as JSON."""
    with open("config/global_config.json", encoding="UTF-8") as f:
        config = json.loads(f.read())
    return config



"""Helper methods used by the module."""

from __future__ import annotations

import inspect
import re
import subprocess
from contextlib import suppress
from dataclasses import dataclass
from typing import Any, BinaryIO, Literal

import pytest
from _pytest._io import TerminalWriter
from _pytest.config import Config
from _pytest.python import Function
from _pytest.terminal import TerminalReporter
from adaptavist.const import STATUS_BLOCKED, STATUS_FAIL, STATUS_IN_PROGRESS, STATUS_NOT_EXECUTED, STATUS_PASS

from .constants import COLORMAP


def calc_test_result_status(step_results: list[dict[str, str]]) -> str:
    """
    Calculate overall test result status from list of step results.

    According to Adaptavist test management:

        Blocked & Not Executed -> Blocked
        Blocked & In Progress -> Blocked
        Blocked & Pass -> Blocked
        Blocked & Fail -> Fail

        Fail & Not Executed -> Fail
        Fail & In Progress -> Fail
        Fail & Pass -> Fail

        Pass & Not Executed -> In Progress
        Pass & In Progress -> In Progress

        In Progress & Not Executed -> In Progress
    """
    # map representing status as binary/hex number to be used with & operator
    status_map = {
        STATUS_NOT_EXECUTED: 0xB,  # 1011
        STATUS_PASS: 0x7,  # 0111
        STATUS_IN_PROGRESS: 0x3,  # 0011
        STATUS_BLOCKED: 0x1,  # 0001
        STATUS_FAIL: 0x0,  # 0000
    }
    if not step_results:
        return STATUS_NOT_EXECUTED
    status = 0xF
    for result in step_results:
        status &= status_map[result["status"]]

    return [k for k, v in status_map.items() if v == status][0]


def get_code_base_url() -> str:
    """Get current code base url."""
    code_base = ""
    with suppress(subprocess.CalledProcessError):
        code_base = subprocess.check_output("git config --get remote.origin.url".split()).decode("utf-8").strip()
    return code_base


def get_option_ini(config: Config, value: str) -> Any:
    """Get option from cli or ini configuration."""
    return config.getoption(value) or config.getini(value)


def get_spec(nodeid: str) -> str:
    """Get callspec from item nodeid."""
    tokens = nodeid.split("[", 1)
    return "[" + tokens[1].strip() if len(tokens) > 1 else ""


def get_item_nodeid(item: pytest.Item) -> str:
    """Build item node id."""
    # note: pytest's item.nodeid could be modified by third party, so build a local one here
    if getattr(item, "location", False) and len(item.location) > 2:
        return item.location[0].replace("\\", "/") + "::" + item.location[2].replace(".", "::")
    return ""


def html_row(condition: Literal["passed", "failed", "blocked"], message: str) -> str:
    """Generate an html status row to be displayed in test case results."""
    if not message:
        return ""

    if condition == "passed":
        background_color = "rgb(58, 187, 75)"
        badge_text = STATUS_PASS
    elif condition == "failed":
        background_color = "rgb(223, 47, 54)"
        badge_text = STATUS_FAIL
    else:
        background_color = "rgb(75, 136, 231)"
        badge_text = STATUS_BLOCKED

    return f"<div style='padding: 2pt'><span style='width: auto; margin-right: 4pt; padding: 2pt; border-radius: 4px; background-color: {background_color}; \
            color: white; font-family: monospace; font-size: 10pt; font-weight: bold;'>{badge_text}</span>{message}</div>"


def intersection(list_a: list[Any], list_b: list[Any]) -> list[Any]:
    """Return the intersection of two lists (maintaining the item order of the first list)."""
    return sorted(set(list_a) & set(list_b), key=list_a.index)


def apply_test_case_range(collected_items: dict[str, list[Function]], test_case_range: list[str]) -> dict[str, list[Function]]:
    """Apply test case range(s) by skipping excluded test cases."""

    if not collected_items or not test_case_range:
        return collected_items

    len_of_range = len(test_case_range)
    out_of_range = len_of_range > 0
    i = 0
    for key in list(collected_items):
        if i < len_of_range and key == test_case_range[i]:
            out_of_range = False

        if out_of_range:
            for item in collected_items[key]:
                if not next(item.iter_markers(name="skip"), None):
                    item.add_marker(pytest.mark.skip(reason="skipped as requested"))
            collected_items.pop(key)

        if i + 1 < len_of_range and key == test_case_range[i + 1]:
            out_of_range = True
            i += 2

    return collected_items


def build_terminal_report(
    when: str,
    item: pytest.Function,
    step: int,
    status: Literal["passed", "failed", "skipped", "blocked"] | None = None,
    level: int = 1,
):
    """
    Generate terminal output.

    :param when: The call info ("setup", "call")
    :param item: The item to report
    :param status: The status ("passed", "failed", "skipped", "blocked")
    :param item: The step index to report
    :param level: The stack trace level (1 = the caller's level, 2 = the caller's caller level, 3 = ...)
    """

    terminal_reporter: TerminalReporter | None = item.config.pluginmanager.getplugin("terminalreporter")
    terminal_writer: TerminalWriter = item.config.get_terminal_writer()

    if terminal_reporter and item.config.option.verbose > 1:
        # extract doc string from source
        (frame, _, line, _, _) = inspect.stack()[level][0:5]
        source_list = inspect.getsourcelines(frame)
        source_code = "".join(source_list[0][line - source_list[1] :])
        docs = re.findall(r"^[\s]*\"\"\"(.*?)\"\"\"", source_code, re.DOTALL | re.MULTILINE | re.IGNORECASE)
        doc_string = inspect.cleandoc(docs[0]) if docs else ""

        if when == "setup":
            terminal_reporter.write_sep("-", "Step " + str(step), bold=True)
            terminal_reporter.write(doc_string + ("\n" if doc_string else ""))
        elif when == "call" and status:
            fill = terminal_writer.fullwidth - terminal_writer.width_of_current_line - 1
            terminal_reporter.write_line(status.upper().rjust(fill), **COLORMAP[status])


@dataclass
class Attachment:
    """Storage for attachments to be uploaded to Adaptavist."""

    attachment: BinaryIO
    """Content of the attachment."""

    filename: str
    """Filename of the attachment."""

    step: int = 0
    """Step to attach it to, 0 for the overall test case."""



"""Fixtures for tests."""
from __future__ import annotations

import shutil
from typing import Generator
from unittest.mock import patch

import pytest
from _pytest.config import Config
from adaptavist import Adaptavist

from . import AdaptavistMock, read_global_config, system_test_preconditions

pytest_plugins = ("pytester",)


def pytest_configure(config: Config):
    """Configure pytest."""
    config.addinivalue_line("markers", "system: mark test as system tests. Select system tests with '-m system'")


@pytest.fixture(scope="session", autouse=True)
def create_test_plan(request: pytest.FixtureRequest) -> str | None:
    """Creates a test plan. All system test will link the test cycle with this test plan."""
    if system_test_preconditions() and request.config.option.markexpr != "not system":
        # This should only be used if test is a system test
        config = read_global_config()
        atm = Adaptavist(config["jira_server"], config["jira_username"], config["jira_password"])
        return atm.create_test_plan(config["project_key"], "pytest_adaptavist_system_test")
    return None


@pytest.fixture(autouse=True)
def set_test_plan(monkeypatch: pytest.MonkeyPatch, request: pytest.FixtureRequest, create_test_plan: str | None):
    """Set test plan key for system tests."""
    if request.node.get_closest_marker("system"):
        monkeypatch.setenv("TEST_PLAN_KEY", create_test_plan or "")


@pytest.fixture
def adaptavist(pytester: pytest.Pytester) -> Generator[Adaptavist, None, None]:
    """Establish connection to Adaptavist."""
    pytester.copy_example("config/global_config.json")
    pytester.mkdir("config")
    shutil.move("global_config.json", "config/global_config.json")
    config = read_global_config()
    yield Adaptavist(config["jira_server"], config["jira_username"], config["jira_password"])


@pytest.fixture(name="test_run")
def create_test_run(monkeypatch: pytest.MonkeyPatch, adaptavist: Adaptavist) -> Generator[str, None, None]:
    """Create a new test run."""
    config = read_global_config()
    if test_run := adaptavist.create_test_run(config["project_key"], "pytest_system_tests"):
        monkeypatch.setenv("TEST_RUN_KEY", test_run)
        yield test_run


@pytest.fixture
def configure(pytester: pytest.Pytester):
    """Configure environment for unit tests."""
    pytester.mkdir("config")
    with open("config/global_config.json", "w", encoding="utf8") as file:
        file.write('{"jira_server": "https://jira.test", "project_key": "TEST", "test_run_key":"TEST-C1"}')


@pytest.fixture
def valid_user() -> Generator[None, None, None]:
    """Mark user as always valid."""
    with patch("pytest_adaptavist.atm_user_is_valid", return_value=True):
        yield


@pytest.fixture
def adaptavist_mock(valid_user: None) -> Generator[AdaptavistMock, None, None]:
    """Patch adaptavist to prevent real I/O."""
    with patch(
        "adaptavist.Adaptavist.get_test_result",
        return_value={"scriptResults": [{"status": "Pass", "index": "0"}], "status": "Pass"},
    ), patch(
        "adaptavist.Adaptavist.get_test_run",
        return_value={"items": [{"testCaseKey": "TEST-T121"}, {"testCaseKey": "TEST-T123"}, {"testCaseKey": "TEST-T124"}]},
    ), patch(
        "adaptavist.Adaptavist.get_test_cases", return_value=[{"key": "TEST-T123"}]
    ), patch(
        "adaptavist.Adaptavist.get_test_run_by_name", return_value={"key": "TEST_RUN_TEST"}
    ), patch(
        "adaptavist.Adaptavist.get_test_case", return_value={"name": "TEST-T123", "priority": "Normal"}
    ), patch(
        "adaptavist.Adaptavist.edit_test_case", return_value=True
    ), patch(
        "adaptavist.Adaptavist._delete"
    ), patch(
        "adaptavist.Adaptavist._get"
    ), patch(
        "adaptavist.Adaptavist._post"
    ), patch(
        "adaptavist.Adaptavist._put"
    ), patch(
        "adaptavist.Adaptavist.create_test_result"
    ) as ctr, patch(
        "adaptavist.Adaptavist.edit_test_result_status"
    ) as etrs, patch(
        "adaptavist.Adaptavist.edit_test_script_status"
    ) as etss:
        yield ctr, etrs, etss



""" Test init file of pytest_adaptavist"""
import pytest


@pytest.mark.usefixtures("adaptavist_mock")
def test_restrict_user(pytester: pytest.Pytester):
    """Test if restrict_user is handles correctly"""
    pytester.makepyfile(
        """
            def test_T1(meta_block):
                with meta_block():
                    with meta_block(1) as mb_1:
                        mb_1.check(True)
        """
    )
    pytester.makeini(
        """
        [pytest]
        restrict_user = abc
    """
    )
    report = pytester.inline_run("--adaptavist")
    assert not report._pluginmanager.get_plugin("_adaptavist").enabled  # pylint: disable=protected-access



"""Process reporting of tests or test steps."""

from __future__ import annotations

import sys
import signal
from datetime import datetime
from enum import IntEnum
from functools import singledispatch
from io import BufferedReader, BytesIO, StringIO
from types import FrameType, TracebackType
from typing import Any, Literal, NoReturn, Tuple

import pytest

from ._helpers import Attachment, build_terminal_report, get_item_nodeid, get_spec, html_row
from ._pytest_adaptavist import PytestAdaptavist


class MetaBlockAborted(Exception):
    """Internal exception used to abort meta block execution."""


class MetaBlock:
    """
    Context Manager class used for processing/reporting single test blocks/steps.

    :param request:
    :param timeout: Timeout in seconds
    :param step: step number as integer starting at 1
    """

    class Action(IntEnum):
        """Action to take, if a test case fails."""

        NONE = 0
        FAIL_CONTEXT = 0
        """If condition fails, collect assumption, set block/test to 'Fail' and continue (just like 'assume')."""
        STOP_CONTEXT = 1
        """If condition fails, skip execution of this block, set it to 'Blocked' and continue with next block."""
        FAIL_METHOD = 2
        """If condition fails, skip execution of this block/test, set it to 'Fail' and continue with next test (just like 'assert')."""
        STOP_METHOD = 3
        """If condition fails, skip execution of this block/test, set it to 'Blocked' and continue with next test."""
        FAIL_SESSION = 4
        """If condition fails, skip execution of this block/test, set it to 'Fail' and block following tests."""
        STOP_SESSION = 5
        """If condition fails, skip execution of this block/test, set it to 'Blocked' and block following tests as well."""
        FAIL_EXIT_SESSION = 6
        """If condition fails, skip execution of this block/test, set it to 'Blocked' and exit session."""
        STOP_EXIT_SESSION = 7
        """If condition fails, skip execution of this block/test, set it to 'Blocked' and exit session."""

    def __init__(
        self,
        request: pytest.FixtureRequest,
        timeout: int,
        action_on_timeout: Action,
        message_on_timeout: str,
        step: int | None = None,
    ):
        fullname = get_item_nodeid(request.node)
        self.item = request.node
        self.items = request.session.items
        self.item_name = self.item.name + ("_" + str(step) if step else "")
        self.step = step
        self.start = datetime.now().timestamp()
        self.stop = datetime.now().timestamp()
        self.timeout = timeout
        self.action_on_timeout = action_on_timeout
        self.message_on_timeout = message_on_timeout
        self.adaptavist: PytestAdaptavist = request.config.pluginmanager.getplugin("_adaptavist")
        self.data: dict[str, Any] = self.adaptavist.test_result_data.setdefault(
            fullname + ("_" + str(step) if step else ""), {"comment": None, "attachment": None}
        )

    @staticmethod
    def _timeout_handler(signum: int, frame: FrameType | None) -> NoReturn:
        """Handle test cases running to long."""
        raise TimeoutError("The test step exceeded its timewindow and timed out")

    def __enter__(self) -> MetaBlock:
        if self.step:
            # level = 2 to get info from outside of this plugin (i.e. caller of 'with metablock(...)')
            build_terminal_report(when="setup", item=self.item, step=self.step, level=2)
        self.start = datetime.now().timestamp()
        self.adaptavist.failed_assumptions_step = []
        if not sys.platform == 'win32':
            signal.signal(signal.SIGALRM, self._timeout_handler)
            signal.alarm(self.timeout)
        return self

    def __exit__(self, exc_type: type, exc_value: Exception, traceback: TracebackType) -> bool:
        if not sys.platform == 'win32':
            signal.alarm(0)
        self.stop = datetime.now().timestamp()
        fullname = get_item_nodeid(self.item)
        if exc_type is TimeoutError:
            self._process_failed_condition(self.action_on_timeout, self.message_on_timeout)
        skip_status = self.item.get_closest_marker("block") or self.item.get_closest_marker("skip")

        # if method was blocked dynamically (during call) an appropriate marker is used
        # to handle the reporting in the same way as for statically blocked methods
        # (status will be reported as "Blocked" with given comment in Adaptavist)
        if not skip_status and (
            exc_type
            and exc_type in (pytest.block.Exception, pytest.skip.Exception)  # type:ignore
            or exc_type in (None, MetaBlockAborted)
            and self.data.get("blocked") is True
        ):
            reason = self.data.get("comment") or (
                str(exc_value).partition("\n")[0]
                if exc_type and exc_type in (pytest.block.Exception, pytest.skip.Exception)  # type:ignore
                else ""
            )

            skip_status = (
                pytest.mark.block(reason=reason)
                if (
                    (exc_type and exc_type is pytest.block.Exception)  # type:ignore
                    or self.data.get("blocked", None) is True
                )
                else pytest.mark.skip(reason=reason)
            )

        # report exceptions
        if exc_type and exc_type is not MetaBlockAborted:
            exc_info = self.adaptavist.build_exception_info(fullname, exc_type, exc_value, traceback)

            if (
                exc_info
                and exc_info not in (self.data.get("comment") or "")
                and (exc_type is not pytest.skip.Exception)
                and not skip_status
            ):
                self.data["comment"] = "".join((self.data.get("comment", None) or "", html_row("failed", exc_info)))

        passed = not exc_type and (
            len(self.adaptavist.failed_assumptions_step) <= len(getattr(pytest, "_failed_assumptions", [])[:])
        )
        status: Literal["passed", "failed", "skipped", "blocked"] = (
            ("passed" if passed else "failed")
            if not skip_status
            else ("blocked" if (skip_status.name == "block" or self.data.get("blocked")) else "skipped")
        )

        # custom item callback
        prefix = (
            getattr(self.item.config, "workerinput", {}).get("workerid")
            if getattr(self.item.config, "workerinput", {}).get("options", {}).get("dist") == "each"
            else None
        )
        getattr(self.item, "meta_block_cb", lambda **kwargs: None)(
            signature="_".join(filter(None, (prefix, self.item.name, str(self.step) if self.step else "x"))), status=status
        )

        if self.step:
            build_terminal_report(
                when="call", item=self.item, status=status, step=self.step, level=2
            )  # level = 2 to get info from outside of this plugin (i.e. caller of 'with metablock(...)'))

        # adjust parent's test result status if necessary (needed for makereport call later)
        if self.adaptavist.test_result_data[fullname].get("blocked") is True and not passed and not skip_status:
            self.adaptavist.test_result_data[fullname]["blocked"] = None
        elif self.data.get("blocked") is True:
            self.adaptavist.test_result_data[fullname]["blocked"] = True

        if not getattr(self.item.config.option, "adaptavist", False):
            # adaptavist reporting disabled: no need to proceed here
            return exc_type is MetaBlockAborted  # suppress MetaBlockAborted exception

        marker = self.item.get_closest_marker("testcase")
        if marker is not None:
            test_case_key = marker.kwargs["test_case_key"]
            try:
                test_step_key = marker.kwargs["test_step_key"]
            except KeyError:
                test_step_key = None

            if test_step_key or not self.step:
                # if it's a test step method, we should not be here
                # if it's the test case context, we can return here as well
                # pytest_runtest_makereport takes care about reporting in both cases
                return exc_type is MetaBlockAborted  # suppress MetaBlockAborted exception

            specs = get_spec(get_item_nodeid(self.item))
            self.adaptavist.create_report(
                test_case_key, self.step, self.stop - self.start, skip_status, passed, self.data, specs
            )

        self.data["done"] = True  # tell pytest_runtest_makereport that this item has been processed already

        return exc_type is MetaBlockAborted  # suppress MetaBlockAborted exception

    def check(self, condition: bool, message: str | None = None, action_on_fail: Action = Action.NONE, **kwargs: Any):
        """
        Check given condition.

        :param condition: The condition to be checked
        :param message: The info test in case of failed condition
        :param action_on_fail: Action in case of failed condition (default: continue, just like 'assume')
        :key attachment: The attachment as filepath name or file-like object
        :key filename: The optional filename
        :key message_on_fail: The info test in case of failed condition (same as message)
        :key message_on_pass: The info test in case of passed condition
        :key description: Optional details about test results (f.e. can be a html table or more)
        """
        attachment = kwargs.pop("attachment", None)
        filename = kwargs.pop("filename", None)
        description = kwargs.pop("description", None)
        message_on_fail = kwargs.pop("message_on_fail", None) or message or ""
        message_on_pass = kwargs.pop("message_on_pass", "")
        if kwargs:
            raise SyntaxWarning(f"Unknown arguments: {kwargs}")

        if attachment and self.adaptavist.enabled:
            if not self.data.get("attachment_test_case"):
                self.data["attachment_test_case"] = []

            if not self.data.get("attachment_test_step"):
                self.data["attachment_test_step"] = []

            content, name = _read_attachment(attachment)
            if self.step:
                self.data["attachment_test_step"].append(
                    Attachment(content, filename=filename or name or "", step=self.step or 0)
                )
            else:
                self.data["attachment_test_case"].append(
                    Attachment(content, filename=filename or name or "", step=self.step or 0)
                )

        if not condition and message_on_fail:
            self.data["comment"] = "".join((self.data.get("comment", "") or "", html_row("failed", message_on_fail)))
        elif condition and message_on_pass:
            self.data["comment"] = "".join((self.data.get("comment", "") or "", html_row("passed", message_on_pass)))

        if description:
            self.data["description"] = "<br>".join((self.data.get("description", ""), description))

        # custom item callback
        prefix = (
            getattr(self.item.config, "workerinput", {}).get("workerid")
            if getattr(self.item.config, "workerinput", {}).get("options", {}).get("dist") == "each"
            else None
        )
        self.__dict__["numchecks"] = self.__dict__.get("numchecks", 0) + 1
        signature = "_".join(
            filter(None, (prefix, self.item.name, str(self.step) if self.step else "x", str(self.__dict__["numchecks"])))
        )
        getattr(self.item, "meta_block_condition_cb", lambda **kwargs: None)(
            signature=signature, condition=condition, reference=message_on_pass if condition else message_on_fail
        )

        if condition:
            return

        self._process_failed_condition(action_on_fail, message_on_fail)

    def _process_failed_condition(self, action_on_fail: Action, message_on_fail: str):
        """Process failed condition depending on action_on_fail."""
        fullname = get_item_nodeid(self.item)
        if action_on_fail == self.Action.FAIL_METHOD:
            # FAIL_METHOD: skip execution of this block/test, set it to 'Fail' and continue with next test
            assert False, message_on_fail
        elif action_on_fail == self.Action.STOP_CONTEXT:
            # STOP_CONTEXT: skip execution of this block, set it to 'Blocked' and continue with next block
            self.data["blocked"] = True
            raise MetaBlockAborted()
        elif action_on_fail == self.Action.STOP_METHOD:
            # STOP_METHOD: skip execution of this block/test, set it to 'Blocked' and continue with next test
            self.data["blocked"] = True
            pytest.skip(msg=f"Blocked. {self.item_name} failed: {message_on_fail}")
        elif action_on_fail == self.Action.STOP_SESSION:
            # STOP_SESSION: skip execution of this block/test, set it to 'Blocked' and block following tests as well
            for item in self.items:
                item.add_marker("block")
                self.adaptavist.test_result_data[fullname]["blocked"] = True
                self.adaptavist.test_result_data[fullname]["comment"] = f"Blocked. {self.item_name} failed: {message_on_fail}"
            pytest.block(msg=message_on_fail)  # type:ignore
        elif action_on_fail == self.Action.FAIL_SESSION:
            # FAIL_SESSION: skip execution of this block/test, set it to 'Fail' and block following tests
            for item in self.items:
                if item.name not in self.item.name:
                    item.add_marker("block")
                    self.adaptavist.test_result_data[fullname]["blocked"] = True
                    self.adaptavist.test_result_data[fullname][
                        "comment"
                    ] = f"Blocked. {self.item_name} failed: {message_on_fail}"
            assert False, message_on_fail
        elif action_on_fail == self.Action.STOP_EXIT_SESSION:
            # EXIT_SESSION: skip execution of this block/test, set it to 'Blocked' and exit session
            self.item.add_marker("block")
            pytest.exit(msg=f"Exiting pytest. {self.item_name} failed: {message_on_fail}", returncode=1)
        elif action_on_fail == self.Action.FAIL_EXIT_SESSION:
            # EXIT_SESSION: skip execution of this block/test, set it to 'Blocked' and exit session
            pytest.exit(msg=f"Exiting pytest. {self.item_name} failed: {message_on_fail}")
        else:
            # CONTINUE: try to collect failed assumption, set result to 'Fail' and continue
            pytest.assume(expr=False, msg=message_on_fail)  # type:ignore


@singledispatch
def _read_attachment(attachment: Any) -> Tuple[BytesIO, str]:
    """Read content of an attachment."""
    raise TypeError(f"Type {type(attachment)} is not supported for attachments.")


@_read_attachment.register
def _(attachment: str) -> Tuple[BytesIO, str]:
    """Read content of an attachment given with filename."""
    with open(attachment, "rb") as file_pointer:
        return BytesIO(file_pointer.read()), file_pointer.name


@_read_attachment.register  # type: ignore
def _(attachment: BufferedReader) -> Tuple[BytesIO, str]:
    """Read content of an attachment given as file pointer."""
    return BytesIO(attachment.read()), attachment.name


@_read_attachment.register  # type: ignore
def _(attachment: StringIO) -> Tuple[BytesIO, str]:
    """Read content of an attachment given as in-memory text buffer."""
    if not hasattr(attachment, "name"):
        raise ValueError("Please give your attachment a name.")
    return BytesIO(bytes(attachment.getvalue(), encoding="utf-8")), attachment.name



"""Test compatibility with pytest-xdist."""
from unittest.mock import patch

import pytest


class TestXdistUnit:
    """Test compatibility with pytest-xdist on unit test level."""

    @pytest.mark.usefixtures("adaptavist_mock")
    def test_xdist_handling(self, pytester: pytest.Pytester):
        """Test coexistence with xdist."""
        with patch("_pytest.config.PytestPluginManager.hasplugin", return_value=True):
            config = pytester.parseconfigure("--adaptavist")
            assert config.pluginmanager.getplugin("_xdist_adaptavist")
        with patch("_pytest.config.PytestPluginManager.hasplugin", return_value=False):
            config = pytester.parseconfigure("--adaptavist")
            assert not config.pluginmanager.getplugin("_xdist_adaptavist")



"""Test compatibility with Adaptavist."""

import getpass
import logging
from unittest.mock import patch

import pytest

from . import AdaptavistMock


@pytest.mark.usefixtures("configure")
class TestAdaptavistUnit:
    """Test compatibility with Adaptavist on unit test level."""

    def test_adaptavist_reporting(self, pytester: pytest.Pytester, adaptavist_mock: AdaptavistMock):
        """Test reporting results to Adaptavist."""
        pytester.makepyfile(
            """
            import pytest

            def test_TEST_T123():
                assert True
        """
        )
        ctr, _, _ = adaptavist_mock
        pytester.runpytest("--adaptavist")
        ctr.assert_called_once_with(
            test_run_key="TEST-C1",
            test_case_key="TEST-T123",
            environment=None,
            executor=getpass.getuser().lower(),
            assignee=getpass.getuser().lower(),
        )

    @pytest.mark.usefixtures("adaptavist_mock")
    def test_unknown_user(self, pytester: pytest.Pytester, caplog: pytest.LogCaptureFixture):
        """Test the correct behavior of an unknown user."""
        pytester.makepyfile(
            """
            import pytest

            def test_TEST_T123():
                assert True
        """
        )
        with patch("pytest_adaptavist.atm_user_is_valid", return_value=False):
            pytester.runpytest("--adaptavist")
            assert caplog.records[-1].funcName == "pytest_configure"
            assert caplog.records[-1].levelno == logging.WARN
            assert (
                caplog.records[-1].msg
                == "Local user '%s' is not known in Jira. Test cases will be reported without an executor!"
            )

    def test_test_case_name_step(self, pytester: pytest.Pytester, adaptavist_mock: AdaptavistMock):
        """Test reporting results to Adaptavist if the step is set in testcase name."""
        pytester.makepyfile(
            """
            import pytest

            def test_TEST_T123_1():
                assert True
        """
        )
        _, _, etss = adaptavist_mock
        pytester.runpytest("--adaptavist")
        etss.assert_called_once_with(
            test_run_key="TEST-C1",
            test_case_key="TEST-T123",
            step=1,
            status="Pass",
            comment="",
            environment=None,
            executor=getpass.getuser().lower(),
            assignee=getpass.getuser().lower(),
        )



"""Test CLI behavior."""
import pytest
from _pytest.config import ExitCode
from _pytest.pytester import LineMatcher


@pytest.mark.usefixtures("configure")
class TestCliUnit:
    """Test CLI behavior on unit test level."""

    def test_adaptavist_disabled(self, pytester: pytest.Pytester):
        """Test absence of adaptavist config if disabled."""
        config = pytester.parseconfig()
        assert not config.getoption("adaptavist")
        assert not config.getini("adaptavist")

    def test_adaptavist_enabled(self, pytester: pytest.Pytester):
        """Test presence of adaptavist config if enabled."""
        config = pytester.parseconfig("--adaptavist")
        assert config.getoption("adaptavist")

        pytester.makeini(
            """
            [pytest]
            adaptavist = 1
        """
        )
        config = pytester.parseconfig()
        assert config.getini("adaptavist")

    def test_cycle_info_urls(self, pytester: pytest.Pytester):
        """Test that the cycle information urls are build correctly."""
        pytester.makepyfile(
            """
            import pytest

            def test_TEST_T123():
                assert True
        """
        )
        report = pytester.runpytest("--adaptavist")
        matcher = LineMatcher(report.outlines)
        matcher.fnmatch_lines(
            [
                "traceability:  https://jira.test/secure/Tests.jspa#/reports/traceability/report/view?tql=testResult.projectKey%20IN%20%28%22TEST%22%29%20AND%20testRun.key%20IN%20%28%22TEST-C1%22%29%20AND%20testRun.onlyLastTestResult%20IS%20true&jql=&title=REPORTS.TRACEABILITY_REPORT.TITLE&traceabilityReportOption=COVERAGE_TEST_CASES&traceabilityTreeOption=COVERAGE_TEST_CASES&traceabilityMatrixOption=COVERAGE_TEST_CASES&period=MONTH&scorecardOption=EXECUTION_RESULTS",
                "test_summary:  https://jira.test/secure/Tests.jspa#/reports/testresults/board/view?tql=testResult.projectKey%20IN%20%28%22TEST%22%29%20AND%20testRun.key%20IN%20%28%22TEST-C1%22%29%20AND%20testRun.onlyLastTestResult%20IS%20true&jql=&title=REPORTS.TEST_RESULTS_BOARD.TITLE&traceabilityReportOption=COVERAGE_TEST_CASES&traceabilityTreeOption=COVERAGE_TEST_CASES&traceabilityMatrixOption=COVERAGE_TEST_CASES&period=MONTH&scorecardOption=EXECUTION_RESULTS",
                "score_matrix:  https://jira.test/secure/Tests.jspa#/reports/testresults/scorecard/coverage/view?tql=testResult.projectKey%20IN%20%28%22TEST%22%29%20AND%20testRun.key%20IN%20%28%22TEST-C1%22%29%20AND%20testRun.onlyLastTestResult%20IS%20true&jql=&title=REPORTS.TEST_RESULTS_SCORECARD_BY_COVERAGE.TITLE&traceabilityReportOption=COVERAGE_TEST_CASES&traceabilityTreeOption=COVERAGE_TEST_CASES&traceabilityMatrixOption=COVERAGE_TEST_CASES&period=MONTH&scorecardOption=EXECUTION_RESULTS",
            ]
        )

    def test_invalid_branch(self, pytester: pytest.Pytester, monkeypatch: pytest.MonkeyPatch):
        """Test the correct behavior of an invalid branch."""
        pytester.makepyfile(
            """
            import pytest

            def test_TEST_T123():
                assert True
        """
        )
        monkeypatch.setenv("GIT_BRANCH", "test")
        report = pytester.runpytest("--adaptavist", "--restrict-branch")
        assert report.ret == ExitCode.INTERNAL_ERROR
        report = pytester.runpytest("--restrict-branch")
        assert report.ret == ExitCode.OK
        report = pytester.runpytest("--adaptavist", "--restrict-branch", "--restrict-branch-name=test")
        assert report.ret == ExitCode.OK
        monkeypatch.setenv("GIT_BRANCH", "origin/master")
        report = pytester.runpytest("--adaptavist", "--restrict-branch")
        assert report.ret == ExitCode.OK



"""Test meta block functionality."""
from unittest.mock import patch

import pytest

from . import AdaptavistMock


@pytest.mark.usefixtures("configure")
class TestMetaBlockUnit:
    """Test meta block functionality on unit test level."""

    def test_message_on_fail(self, pytester: pytest.Pytester, adaptavist_mock: AdaptavistMock):
        """Test the correct usage of the message_on_fail parameter of meta_block check."""

        # Test message_on_fail for a failing test case
        pytester.makepyfile(
            """
            import pytest

            def test_TEST_T123(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(False, message_on_fail="We want to see this message")
        """
        )
        _, _, etss = adaptavist_mock
        pytester.runpytest("--adaptavist")
        assert "We want to see this message" in etss.call_args.kwargs["comment"]

        # Test message_on_fail for a passing test case
        pytester.makepyfile(
            """
            import pytest

            def test_TEST_T123(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(True, message_on_fail="We don't want to see this message")
        """
        )
        pytester.runpytest("--adaptavist")
        assert "We don't want to see this message" not in etss.call_args.kwargs["comment"]

    def test_message_on_pass(self, pytester: pytest.Pytester, adaptavist_mock: AdaptavistMock):
        """Test the correct usage of the message_on_pass parameter of meta_block check."""

        # Test message_on_pass for a passing test case
        pytester.makepyfile(
            """
            import pytest

            def test_TEST_T123(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(True, message_on_pass="We want to see this message")
        """
        )
        _, _, etss = adaptavist_mock
        pytester.runpytest("--adaptavist")
        assert "We want to see this message" in etss.call_args.kwargs["comment"]

        # Test message_on_pass for a failing test case
        pytester.makepyfile(
            """
            import pytest

            def test_TEST_T123(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(False, message_on_pass="We don't want to see this message")
        """
        )
        pytester.runpytest("--adaptavist")
        assert "We don't want to see this message" not in etss.call_args.kwargs["comment"]

    def test_description(self, pytester: pytest.Pytester, adaptavist_mock: AdaptavistMock):
        """Test the correct usage of the description parameter of meta_block check."""
        pytester.makepyfile(
            """
            import pytest

            def test_TEST_T123(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(True, description="This should be displayed twice")
                    mb_1.check(False, description="This should be displayed twice")
        """
        )
        _, _, etss = adaptavist_mock
        pytester.runpytest("--adaptavist")
        assert etss.call_args.kwargs["comment"].count("This should be displayed twice") == 2

    @pytest.mark.usefixtures("adaptavist_mock")
    def test_attachment(self, pytester: pytest.Pytester):
        """Test the correct usage of the attachment parameter."""
        pytester.maketxtfile(first_file="foo")

        # Test attachment for a passing test case
        pytester.makepyfile(
            """
            import pytest

            def test_TEST_T123(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(True, attachment="first_file.txt")
        """
        )
        with patch("adaptavist.Adaptavist.add_test_script_attachment") as atsa:
            pytester.runpytest("--adaptavist")
            assert "first_file.txt" in atsa.call_args.kwargs["filename"]

        # Test attachment for a failing test case
        pytester.makepyfile(
            """
            import pytest

            def test_TEST_T123(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(False, attachment="first_file.txt")
        """
        )
        with patch("adaptavist.Adaptavist.add_test_script_attachment") as atsa:
            pytester.runpytest("--adaptavist")
            assert "first_file.txt" in atsa.call_args.kwargs["filename"]

        # Test attaching with a file handle
        pytester.makepyfile(
            """
            import pytest

            def test_TEST_T123(meta_block):
                with meta_block(1) as mb_1, open("first_file.txt", "rb") as fh:
                    mb_1.check(False, attachment=fh)
        """
        )
        with patch("adaptavist.Adaptavist.add_test_script_attachment") as atsa:
            pytester.runpytest("--adaptavist")
            assert "first_file.txt" in atsa.call_args.kwargs["filename"]

        # Test attaching a StringIO object
        pytester.makepyfile(
            """
            import pytest
            from io import StringIO

            def test_TEST_T123(meta_block):
                with meta_block(1) as mb_1:
                    attachment = StringIO()
                    attachment.name = "first_file.txt"
                    mb_1.check(False, attachment=attachment)
        """
        )
        with patch("adaptavist.Adaptavist.add_test_script_attachment") as atsa:
            pytester.runpytest("--adaptavist")
            assert "first_file.txt" in atsa.call_args.kwargs["filename"]

    @pytest.mark.usefixtures("adaptavist_mock")
    def test_description_of_test_steps_printed(self, pytester: pytest.Pytester):
        """Test if the description of testcases is printed, if pytest is started with high verbosity."""
        pytester.makepyfile(
            '''
            import pytest

            def test_TEST_T123(meta_block):
                with meta_block(1) as mb_1:
                    """Description of test step 1"""
                    mb_1.check(True)
        '''
        )

        # Low verbosity
        report = pytester.runpytest("--adaptavist")
        assert "Description of test step 1" not in report.outlines

        # High verbosity
        report = pytester.runpytest("--adaptavist", "-vv")
        assert "Description of test step 1" in report.outlines

    def test_adaptavist_call_metablock(self, pytester: pytest.Pytester, adaptavist_mock: AdaptavistMock):
        """Test if all meta blocks are reported to adaptavist."""
        pytester.makepyfile(
            """
            import pytest

            def test_TEST_T123(meta_block):
                with meta_block(1):
                    assert True
                with meta_block(2):
                    assert True
                with meta_block(3):
                    assert False
        """
        )
        _, etrs, _ = adaptavist_mock
        pytester.runpytest("--adaptavist")
        assert etrs.call_count == 4  # 3 meta blocks and 1 overall result

    @pytest.mark.usefixtures("adaptavist_mock")
    def test_meta_block_timeout(self, pytester: pytest.Pytester):
        """Test if a meta block is timed out."""
        pytester.makepyfile(
            """
            from time import sleep
            import pytest

            def test_TEST_T123(meta_block):
                with meta_block(1, 1):
                    sleep(2)
                    assert True
        """
        )
        pytester.runpytest("--adaptavist").assert_outcomes(skipped=1)

    @pytest.mark.usefixtures("adaptavist_mock")
    def test_meta_block_timeout_fail(self, pytester: pytest.Pytester):
        """Test if a meta block is timed out with fail action."""
        pytester.makepyfile(
            """
            from time import sleep
            import pytest

            def test_TEST_T123(meta_block):
                with meta_block(1, timeout=1, action_on_timeout=0):
                    sleep(2)
                    assert True
        """
        )
        pytester.runpytest("--adaptavist").assert_outcomes(failed=1)

    @pytest.mark.usefixtures("adaptavist_mock")
    def test_meta_block_check_unknown_arguments(self, pytester: pytest.Pytester):
        """Test if meta_block handles unknown keyword arguments correctly."""
        pytester.makepyfile(
            """
            import pytest

            def test_TEST_T123(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(True, unknown_kwargs=123)
        """
        )
        report = pytester.runpytest("--adaptavist")
        assert any(("Unknown arguments: {'unknown_kwargs': 123}" in x for x in report.outlines))

    def test_meta_block_assume(self, pytester: pytest.Pytester, adaptavist_mock: AdaptavistMock):
        """Test if meta_block is using assume correctly. Step 2 must be executed even if step 1 fails."""
        pytester.makepyfile(
            """
            import pytest

            def test_TEST_T123(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(False)
                with meta_block(2) as mb_2:
                    mb_2.check(True)
        """
        )
        pytester.runpytest("--adaptavist")
        _, _, etss = adaptavist_mock
        assert etss.call_count == 2

    def test_meta_block_fail_method(self, pytester: pytest.Pytester, adaptavist_mock: AdaptavistMock):
        """Test Action.FAIL_METHOD. Step 2 must not be executed as step 1 will fail."""
        pytester.makepyfile(
            """
            import pytest

            def test_TEST_T123(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(False, action_on_fail=mb_1.Action.FAIL_METHOD)
                with meta_block(2) as mb_2:
                    mb_2.check(True)
        """
        )
        pytester.runpytest("--adaptavist")
        _, _, etss = adaptavist_mock
        assert etss.call_count == 1

    def test_meta_block_stop_context(self, pytester: pytest.Pytester, adaptavist_mock: AdaptavistMock):
        """Test Action.STOP_CONTEXT. The second check in meta_block 1 must not be executed, but step 2 must be executed."""
        pytester.makepyfile(
            """
            import pytest

            def test_TEST_T123(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(False, action_on_fail=mb_1.Action.STOP_CONTEXT)
                    mb_1.check(False, message_on_fail="THIS SHOULD NOT BE DISPLAYED")
                with meta_block(2) as mb_2:
                    mb_2.check(True)
        """
        )
        report = pytester.runpytest("--adaptavist", "-vv")
        _, _, etss = adaptavist_mock
        for call in etss.call_args_list:
            assert "THIS SHOULD NOT BE DISPLAYED" not in call.kwargs["comment"]
        assert etss.call_count == 2
        assert etss.call_args.kwargs["step"] == 2
        assert etss.call_args.kwargs["status"] == "Pass"
        report.assert_outcomes(skipped=1)

    def test_meta_block_stop_method(self, pytester: pytest.Pytester, adaptavist_mock: AdaptavistMock):
        """Test Action.STOP_METHOD. We expect to not see step 2 and the second check of meta_block 1. TEST-T124 must be executed normally."""

        pytester.makepyfile(
            """
            import pytest

            def test_TEST_T123(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(False, action_on_fail=mb_1.Action.STOP_METHOD)
                    mb_1.check(False, message_on_fail="THIS SHOULD NOT BE DISPLAYED")
                with meta_block(2) as mb_2:
                    mb_2.check(True)

            def test_TEST_T124(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(True)
        """
        )
        pytester.runpytest("--adaptavist")
        _, _, etss = adaptavist_mock
        assert etss.call_count == 2
        for call in etss.call_args_list:
            assert "THIS SHOULD NOT BE DISPLAYED" not in call.kwargs["comment"]
            assert call.test_case_key != "TEST-T123" or call.step != 2

    @pytest.mark.usefixtures("adaptavist_mock")
    def test_meta_block_check_stop_session(self, pytester: pytest.Pytester):
        """Test Action.STOP_SESSION. We expect that TEST_T121 is executed. TEST_T123 fails and prevent execution of TEST_T124"""
        pytester.makepyfile(
            """
            import pytest

            def test_TEST_T121(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(True)

            def test_TEST_T123(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(False, action_on_fail=mb_1.Action.STOP_SESSION)

            def test_TEST_T124(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(True)
        """
        )
        outcome = pytester.runpytest("--adaptavist").parseoutcomes()
        assert outcome["passed"] == 1
        assert outcome["blocked"] == 2

    @pytest.mark.usefixtures("adaptavist_mock")
    def test_meta_block_check_fail_session(self, pytester: pytest.Pytester):
        """Test Action.FAIL_SESSION. We expect that TEST_T121 is executed. TEST_T123 fails and fails the whole session."""
        pytester.makepyfile(
            """
            import pytest

            def test_TEST_T121(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(True)

            def test_TEST_T123(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(False, action_on_fail=mb_1.Action.FAIL_SESSION)

            def test_TEST_T124(meta_block):
                with meta_block(1):
                    assert True
        """
        )
        outcome = pytester.runpytest("--adaptavist").parseoutcomes()
        assert outcome["failed"] == 1
        assert outcome["blocked"] == 1
        assert outcome["passed"] == 1

        test_runs = {"items": [{"testCaseKey": "TEST-T123"}, {"testCaseKey": "TEST-T124"}, {"testCaseKey": "TEST-T121"}]}
        with patch("adaptavist.Adaptavist.get_test_run", return_value=test_runs):
            outcome = pytester.runpytest("--adaptavist").parseoutcomes()
            assert outcome["failed"] == 1
            assert outcome["blocked"] == 2

    def test_meta_block_stop_exit_session(self, pytester: pytest.Pytester, adaptavist_mock: AdaptavistMock):
        """Test Action.EXIT_SESSION. We expect that TEST_T123 fails and exits the whole session."""
        pytester.makepyfile(
            """
            import pytest

            def test_TEST_T123(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(False, action_on_fail=mb_1.Action.STOP_EXIT_SESSION)
                    mb_1.check(False, message_on_fail="THIS SHOULD NOT BE DISPLAYED")
                with meta_block(2) as mb_2:
                    mb_2.check(True)

            def test_TEST_T124(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(True)
        """
        )
        pytester.runpytest("--adaptavist")
        _, _, etss = adaptavist_mock
        assert etss.call_count == 1
        assert etss.call_args.kwargs["status"] == "Blocked"

    def test_meta_block_fail_exit_session(self, pytester: pytest.Pytester, adaptavist_mock: AdaptavistMock):
        """Test Action.EXIT_SESSION. We expect that TEST_T123 fails and exits the whole session."""
        pytester.makepyfile(
            """
            import pytest

            def test_TEST_T123(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(False, action_on_fail=mb_1.Action.FAIL_EXIT_SESSION)
                    mb_1.check(False, message_on_fail="THIS SHOULD NOT BE DISPLAYED")
                with meta_block(2) as mb_2:
                    mb_2.check(True)

            def test_TEST_T124(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(True)
        """
        )
        pytester.runpytest("--adaptavist")
        _, _, etss = adaptavist_mock
        assert etss.call_count == 1
        assert etss.call_args.kwargs["status"] == "Fail"



"""Test pytest.ini configuration."""
import os
from importlib.metadata import version

import pytest
from adaptavist import Adaptavist

from pytest_adaptavist import PytestAdaptavist
from tests import get_test_values, read_global_config, system_test_preconditions


@pytest.mark.usefixtures("adaptavist_mock")
class TestIniConfigUnit:
    """Test pytest.ini configuration on unit test level."""

    @pytest.mark.parametrize(
        "option",
        [
            "project_key",
            "test_case_keys",
            "test_case_order",
            "test_case_range",
            "test_environment",
            "test_plan_folder",
            "test_plan_suffix",
            "test_plan_key",
            "test_run_folder",
            "test_run_key",
            "test_run_suffix",
        ],
    )
    def test_ini_config_strings(self, pytester: pytest.Pytester, monkeypatch: pytest.MonkeyPatch, option: str):
        """Test that string values in pytest.ini are correctly used and recognized by pytest."""
        monkeypatch.setenv("PYTEST_DISABLE_PLUGIN_AUTOLOAD", "1")
        pytester.makepyfile(
            """
            def test_T1(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(True)
        """
        )
        pytester.makeini(
            f"""
            [pytest]
            {option} = C1
            """
        )

        report = pytester.inline_run("--adaptavist", plugins=["adaptavist", "assume"])
        assert getattr(report._pluginmanager.get_plugin("_adaptavist"), option) in (  # pylint: disable=protected-access
            "C1",
            ["C1"],
        )

        result = pytester.runpytest("--adaptavist", plugins=["adaptavist", "assume"])
        assert "warnings" not in result.parseoutcomes()

        monkeypatch.setenv(option, "C2")
        report = pytester.inline_run("--adaptavist", plugins=["adaptavist", "assume"])
        assert getattr(report._pluginmanager.get_plugin("_adaptavist"), option) in (  # pylint: disable=protected-access
            "C2",
            ["C2"],
        )

    @pytest.mark.xfail(
        version("adaptavist") < "2.3.1",
        reason="Session usage was introduced in version 2.3.1.",
    )
    def test_jira_settings(self, pytester: pytest.Pytester) -> None:
        """Test that jira settings in pytest.ini are correctly used and recognized by pytest."""
        pytester.makepyfile(
            """
            def test_T1(meta_block):
            with meta_block(1) as mb_1:
            mb_1.check(True)
        """
        )
        pytester.makeini(
            """
            [pytest]
            jira_server = https://jira.test
            jira_username = username
            jira_password = password
        """
        )
        report = pytester.inline_run("--adaptavist", plugins=["adaptavist", "assume"])
        adaptavist: PytestAdaptavist = report._pluginmanager.get_plugin("_adaptavist")  # pylint: disable=protected-access
        assert adaptavist.adaptavist.jira_server == "https://jira.test"
        assert adaptavist.adaptavist._session.auth.username == "username"  # pylint: disable=protected-access
        assert adaptavist.adaptavist._session.auth.password == "password"  # pylint: disable=protected-access


@pytest.mark.system
@pytest.mark.skipif(not system_test_preconditions(), reason="Preconditions for system tests not met. Please see README.md")
class TestIniConfigSystem:
    """Test pytest.ini configuration on system test level."""

    def test_T1(self, pytester: pytest.Pytester, adaptavist: Adaptavist):
        """Test passing a test."""
        pytester.makepyfile(
            """
            def test_T1(meta_block):
                with meta_block():
                    with meta_block(1) as mb_1:
                        mb_1.check(True)
        """
        )
        config = read_global_config()
        pytester.makeini(
            f"""
            [pytest]
            project_key = {config["project_key"]}
            jira_server = {config["jira_server"]}
            jira_username = {config["jira_username"]}
            jira_password = {config["jira_password"]}
        """
        )
        os.remove("config/global_config.json")
        report = pytester.inline_run("--adaptavist")
        test_run_key, test_name = get_test_values(report)
        test_result = adaptavist.get_test_result(test_run_key, test_name)
        assert test_result["status"] == "Pass"
        assert test_result["scriptResults"][0]["status"] == "Pass"



""" Test Adaptavist test management configuration."""
import pytest
from _pytest.config import ExitCode

from pytest_adaptavist._atm_configuration import ATMConfiguration


def test_get(monkeypatch: pytest.MonkeyPatch):
    """Test atm get function. Config dictionary is preferred over OS environment."""
    atm_config = ATMConfiguration()
    atm_config.global_config["cfg_test_variable"] = "correct source"
    monkeypatch.setenv("cfg_test_variable", "wrong source")
    assert atm_config.get("cfg_test_variable") == "correct source"

    atm_config.global_config["cfg_test"] = "test_cfg"
    assert atm_config.get("test") == "test_cfg"


def test_get_environ(monkeypatch: pytest.MonkeyPatch):
    """Test that an OS environment variable is returned if no config is set in dictionary."""
    atm_config = ATMConfiguration()
    monkeypatch.setenv("test_variable", "variable from environment")
    assert atm_config.get("test_variable") == "variable from environment"

    monkeypatch.setenv("UPPER_TEST_VARIABLE", "upper case variable")
    assert atm_config.get("upper_test_variable") == "upper case variable"


@pytest.mark.parametrize(
    "input_values, output_values",
    [
        ("true", True),
        ("1", True),
        ("yes", True),
        ("True", True),
        ("Yes", True),
        (1, True),
        (1000, True),
        ("false", False),
        ("0", False),
        ("no", False),
        ("False", False),
        ("No", False),
        (0, False),
    ],
)
def test_get_bool(input_values, output_values):
    """Test that the get_bool function return correct boolean values for different input values. It tests strings and integers input."""
    atm_config = ATMConfiguration()
    atm_config.global_config["test_bool"] = input_values
    assert atm_config.get_bool("test_bool") is output_values


def test_get_bool_exception():
    """Test that an exception is raised if get_bool can't convert it to a valid boolean value"""
    atm_config = ATMConfiguration()
    atm_config.global_config["test_bool"] = []
    with pytest.raises(ValueError):
        atm_config.get_bool("test_bool")


def test_atm_no_json_file(pytester: pytest.Pytester):
    """Test if atm configuration will fail if there is no valid json found at ./config/global_config.json"""
    pytester.mkdir("config")
    with open("config/global_config.json", "w", encoding="utf8") as file:
        file.write("This is not valid json")
    report = pytester.runpytest()
    assert report.ret == ExitCode.INTERNAL_ERROR



"""Test connection between pytest and Adaptavist."""
import getpass
import re
from importlib.metadata import version
from io import BytesIO
from unittest.mock import patch

import pytest
from _pytest.assertion.util import running_on_ci
from adaptavist import Adaptavist

from . import AdaptavistMock, get_test_values, system_test_preconditions


@pytest.mark.usefixtures("configure")
class TestPytestAdaptavistUnit:
    """Test connection between pytest and Adaptavist on unit test level."""

    @pytest.mark.usefixtures("adaptavist_mock")
    def test_block_call(self, pytester: pytest.Pytester):
        """Test calling block."""
        pytester.makepyfile(
            """
            import pytest

            def test_dummy():
                pytest.block()
                assert True
        """
        )
        outcome = pytester.runpytest().parseoutcomes()
        assert outcome["blocked"] == 1
        assert "passed" not in outcome

    @pytest.mark.usefixtures("adaptavist_mock")
    def test_default_test_project(self, pytester: pytest.Pytester):
        """Test if a project is set to TEST if not found in markers, testcasename or config."""
        pytester.makepyfile(
            """
            import pytest

            def test_T123(meta_block):
                with meta_block(1):
                    assert True
        """
        )
        with open("config/global_config.json", "w", encoding="utf8") as file:
            file.write('{"test_run_key":"TEST-C1"}')
        hook_record = pytester.inline_run("--adaptavist")
        assert hook_record.matchreport().head_line == "test_T123"

    @pytest.mark.usefixtures("adaptavist_mock")
    def test_skip_no_test_case_methods(self, pytester: pytest.Pytester):
        """Test if a test method which is not a valid adaptavist test case is skipped, if 'skip_ntc_methods' is set"""
        pytester.makepyfile(
            """
            import pytest

            def test_not_a_test_case(meta_block):
                with meta_block(1):
                    assert True
        """
        )
        with open("config/global_config.json", "w", encoding="utf8") as file:
            file.write('{"skip_ntc_methods": true}')
        pytester.runpytest("--adaptavist").assert_outcomes(skipped=1)

    def test_early_return_on_no_config(self, pytester: pytest.Pytester, adaptavist_mock: AdaptavistMock):
        """Test the early return in create_report if config is not valid."""
        pytester.makepyfile(
            """
            import pytest

            def test_T123(meta_block):
                with meta_block(1):
                    assert True
        """
        )
        with open("config/global_config.json", "w", encoding="utf8") as file:
            file.write("{}")
        pytester.runpytest("--adaptavist")
        ctr, etrs, etss = adaptavist_mock
        assert ctr.call_count == 0
        assert etrs.call_count == 0
        assert etss.call_count == 0

    @pytest.mark.xfail(
        version("adaptavist") == "2.0.0",
        reason="As long as adaptavist package didn't release the constant fix, this test will fail.",
    )
    def test_skipped_test_cases(self, pytester: pytest.Pytester, adaptavist_mock: AdaptavistMock) -> None:
        """Test that a skipped test case is reported as 'Not Executed'."""
        pytester.makepyfile(
            """
            import pytest

            @pytest.mark.skip()
            def test_T123(meta_block):
                with meta_block(1):
                    assert True
        """
        )
        with patch("adaptavist.Adaptavist.get_test_result", return_value={"scriptResults": [{"index": "0"}]}):
            pytester.runpytest("--adaptavist")
        _, etrs, _ = adaptavist_mock
        assert etrs.call_count == 1
        assert etrs.call_args.kwargs["status"] == "Not Executed"

    def test_test_case_reporting(self, pytester: pytest.Pytester, adaptavist_mock: AdaptavistMock):
        """Test that a testcase is reported correctly. The steps must not reported in this test."""
        pytester.makepyfile(
            """
            import pytest

            def test_T123(meta_block):
                with meta_block() as mb:
                    mb.check(True)
        """
        )
        with patch("adaptavist.Adaptavist.get_test_result", return_value={"scriptResults": [{"index": "0"}]}):
            pytester.runpytest("--adaptavist")
            _, etrs, _ = adaptavist_mock
            assert etrs.call_count == 1
            assert etrs.call_args.kwargs["status"] == "Pass"

            pytester.makepyfile(
                """
                import pytest

                def test_T123(meta_block):
                    with meta_block() as mb:
                        mb.check(False)
            """
            )
            pytester.runpytest("--adaptavist")
            assert etrs.call_count == 2
            assert etrs.call_args.kwargs["status"] == "Fail"

    @pytest.mark.usefixtures("adaptavist_mock")
    def test_result_attachment(self, pytester: pytest.Pytester):
        """Test that an attachment is correctly attached to the testcase (not to the step)."""
        pytester.maketxtfile(first_file="foo")
        pytester.makepyfile(
            """
            import pytest

            def test_T123(meta_block):
                with meta_block() as mb:
                    mb.check(True, attachment="first_file.txt", filename="test.txt")
        """
        )
        with patch("adaptavist.Adaptavist.add_test_result_attachment") as atra:
            pytester.runpytest("--adaptavist")
        assert atra.call_count == 1
        assert isinstance(atra.call_args.kwargs["attachment"], BytesIO)
        assert atra.call_args.kwargs["filename"] == "test.txt"

    @pytest.mark.usefixtures("adaptavist_mock")
    def test_skipped_test_cases_keys(self, pytester: pytest.Pytester):
        """Test that testcases which are not defined in test_case_keys are skipped."""
        pytester.makepyfile(
            """
            import pytest

            def test_T125(meta_block):
                with meta_block() as mb:
                    mb.check(False)

            def test_T123(meta_block):
                with meta_block() as mb:
                    mb.check(True)

            def test_T124(meta_block):
                with meta_block() as mb:
                    mb.check(True)
        """
        )
        # Test that test cases skipped if append-to-cycle is off and test_case_keys are set
        with open("config/global_config.json", "w", encoding="utf8") as file:
            file.write('{"project_key": "TEST", "test_run_key":"TEST-C1", "test_case_keys": ["TEST-T123"]}')
        pytester.runpytest("--adaptavist").assert_outcomes(passed=1, skipped=2)

        # Test that test cases which are not defined in test_case_keys are skipped if append-to-cycle is on
        with open("config/global_config.json", "w", encoding="utf8") as file:
            file.write('{"project_key": "TEST", "test_run_key":"TEST-C1", "test_case_keys": ["TEST-T125"]}')
        pytester.runpytest("--adaptavist", "--append-to-cycle").assert_outcomes(failed=1, skipped=2)

        # Test that test cases run if append-to-cycle is on and test_case_keys are not set
        with open("config/global_config.json", "w", encoding="utf8") as file:
            file.write('{"project_key": "TEST", "test_run_key":"TEST-C1", "test_case_keys": []}')
        pytester.runpytest("--adaptavist", "--append-to-cycle").assert_outcomes(passed=2, failed=1)

    @pytest.mark.usefixtures("adaptavist_mock")
    def test_xfail(self, pytester: pytest.Pytester):
        """Test that xfail is handled properly."""
        pytester.makepyfile(
            """
            import pytest

            @pytest.mark.xfail
            def test_T123(meta_block):
                assert False
        """
        )
        pytester.runpytest("--adaptavist").assert_outcomes(xfailed=1)

    @pytest.mark.usefixtures("adaptavist_mock")
    def test_correct_stacktrace(self, pytester: pytest.Pytester):
        """Test that the correct stack trace is printed."""
        pytester.makepyfile(
            """
                def test_a(meta_block):
                    with meta_block(1) as mb_1:
                        mb_1.check(not True)

                def test_b(meta_block):
                    with meta_block(1) as mb_1:
                        mb_1.check(False)
                    with meta_block(2) as mb_2:
                        mb_2.check(not not False)
        """
        )
        outcome = pytester.runpytest()
        regex = re.findall("\\(not True", str(outcome.outlines).replace("'", "").replace("[", "").replace("]", ""))
        assert len(regex) == 2 if running_on_ci() else 1
        regex = re.findall("\\(False", str(outcome.outlines).replace("'", "").replace("[", "").replace("]", ""))
        assert len(regex) == 2 if running_on_ci() else 1
        regex = re.findall("\\(not not False", str(outcome.outlines).replace("'", "").replace("[", "").replace("]", ""))
        assert len(regex) == 2 if running_on_ci() else 1

    def test_reporting_skipped_test_cases(self, pytester: pytest.Pytester, adaptavist_mock: AdaptavistMock):
        """Don't report a test case if it is not in test_case_keys."""
        pytester.makepyfile(
            """
            import pytest

            class TestClass():
                def test_T121(self, meta_block):
                    pass

                def test_T123(self, meta_block):
                    pass
        """
        )
        with open("config/global_config.json", "w", encoding="utf8") as file:
            file.write('{"test_case_keys": ["TEST-T123"]}')
        _, etrs, _ = adaptavist_mock
        pytester.runpytest("--adaptavist")
        assert etrs.call_args_list[0].kwargs["test_case_key"] == "TEST-T123"
        assert etrs.call_count == 1

    @pytest.mark.filterwarnings("default")
    @pytest.mark.usefixtures("adaptavist_mock")
    def test_deprecated_options(self, pytester: pytest.Pytester):
        """Test deprecated options."""
        pytester.makepyfile(
            """
            def test_dummy():
                assert True
            """
        )
        result = pytester.runpytest("--test_run_name=abc", "--adaptavist")
        assert any(
            "PytestDeprecationWarning: test_run_name is deprecated. Please use --test-cycle-name" in line
            for line in result.outlines
        )

        result = pytester.runpytest("--test_plan_name=abc", "--adaptavist")
        assert any(
            "PytestDeprecationWarning: test_plan_name is deprecated. Please use --test-plan-name" in line
            for line in result.outlines
        )

    @pytest.mark.usefixtures("adaptavist_mock")
    def test_test_run_name(self, pytester: pytest.Pytester):
        """Test that test_run_name template is working."""
        with patch("adaptavist.Adaptavist.create_test_run", return_value="TEST-C123") as ctr, patch(
            "adaptavist.Adaptavist.get_test_run_by_name", return_value={}
        ):
            pytester.makepyfile(
                """
                import pytest

                class TestClass():
                    def test_T121(self, meta_block):
                        pass

                    def test_T123(self, meta_block):
                        pass
            """
            )
            with open("config/global_config.json", "w", encoding="utf8") as file:
                file.write('{"jira_server": "https://jira.test", "project_key": "TEST"}')

            pytester.runpytest("--adaptavist")
            assert "TEST test run" in ctr.call_args_list[0][1]["test_run_name"]

            pytester.makeini(
                """
            [pytest]
            test_run_name = Change test_run_name %(project_key)
            """
            )
            pytester.runpytest("--adaptavist")
            assert ctr.call_args[1]["test_run_name"] == "Change test_run_name TEST"

    @pytest.mark.usefixtures("adaptavist_mock")
    def test_test_plan_name_template(self, pytester: pytest.Pytester):
        """Test that test_run_name template is working."""
        with patch("adaptavist.Adaptavist.create_test_plan") as ctp, patch(
            "adaptavist.Adaptavist.get_test_plans", return_value={}
        ):
            pytester.makepyfile(
                """
                import pytest

                class TestClass():
                    def test_T121(self, meta_block):
                        pass

                    def test_T123(self, meta_block):
                        pass
            """
            )
            with open("config/global_config.json", "w", encoding="utf8") as file:
                file.write('{"jira_server": "https://jira.test", "project_key": "TEST", "test_plan_suffix": "suffix"}')

            pytester.runpytest("--adaptavist")
            assert "TEST suffix" in ctp.call_args_list[0][1]["test_plan_name"]

            pytester.makeini(
                """
                [pytest]
                test_plan_name = Change test_plan_name %(project_key)
            """
            )
            pytester.runpytest("--adaptavist")
            assert "Change test_plan_name TEST" == ctp.call_args[1]["test_plan_name"]

    @pytest.mark.usefixtures("adaptavist_mock")
    def test_test_run_name_invalid_key(self, pytester: pytest.Pytester):
        """Test that test_run_name template is working."""
        with patch("adaptavist.Adaptavist.create_test_run", return_value="TEST-C123"), patch(
            "adaptavist.Adaptavist.get_test_run_by_name", return_value={}
        ):
            pytester.makepyfile(
                """
                import pytest

                class TestClass():
                    def test_T121(self, meta_block):
                        pass

                    def test_T123(self, meta_block):
                        pass
            """
            )
            with open("config/global_config.json", "w", encoding="utf8") as file:
                file.write('{"jira_server": "https://jira.test", "project_key": "TEST"}')
            pytester.makeini(
                """
                [pytest]
                test_run_name = Change test_run_name %(project_ey)
            """
            )
            outcome = pytester.runpytest("--adaptavist")
            assert outcome.ret == 6
            assert any("project_ey" in line for line in outcome.outlines)


@pytest.mark.system
@pytest.mark.skipif(not system_test_preconditions(), reason="Preconditions for system tests not met. Please see README.md")
class TestPytestAdaptavistSystem:
    """Test connection between pytest and Adaptavist on system test level."""

    def test_T1(self, pytester: pytest.Pytester, adaptavist: Adaptavist):
        """Test passing a test."""
        pytester.makepyfile(
            """
            def test_T1(meta_block):
                with meta_block():
                    with meta_block(1) as mb_1:
                        mb_1.check(True)
        """
        )
        pytester.makeini(
            f"""
            [pytest]
            restrict_user = {getpass.getuser().lower()}
        """
        )
        report = pytester.inline_run("--adaptavist")
        test_run_key, test_name = get_test_values(report)
        test_result = adaptavist.get_test_result(test_run_key, test_name)
        assert test_result["status"] == "Pass"
        assert test_result["scriptResults"][0]["status"] == "Pass"

    def test_T2(self, pytester: pytest.Pytester, adaptavist: Adaptavist):
        """Test failing a test."""
        pytester.makepyfile(
            """
            def test_T2(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(False)
        """
        )
        report = pytester.inline_run("--adaptavist")
        test_run_key, test_name = get_test_values(report)
        test_result = adaptavist.get_test_result(test_run_key, test_name)
        assert test_result["status"] == "Fail"
        assert test_result["scriptResults"][0]["status"] == "Fail"

    def test_T3(self, pytester: pytest.Pytester, adaptavist: Adaptavist):
        """Test reporting steps."""
        pytester.makepyfile(
            """
            def test_T3(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(True)
                with meta_block(2) as mb_2:
                    mb_2.check(False)
        """
        )
        report = pytester.inline_run("--adaptavist")
        test_run_key, test_name = get_test_values(report)
        test_result = adaptavist.get_test_result(test_run_key, test_name)
        assert test_result["status"] == "Fail"
        assert test_result["scriptResults"][0]["status"] == "Pass"
        assert test_result["scriptResults"][1]["status"] == "Fail"

    def test_T4(self, pytester: pytest.Pytester, adaptavist: Adaptavist):
        """Test blocking a step."""
        pytester.makepyfile(
            """
            import pytest

            def test_T4(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(True)
                with meta_block(2) as mb_2:
                    pytest.block("Testing block")  # type: ignore
                    mb_2.check(False)
        """
        )
        report = pytester.inline_run("--adaptavist")
        test_run_key, test_name = get_test_values(report)
        test_result = adaptavist.get_test_result(test_run_key, test_name)
        assert test_result["status"] == "Blocked"
        assert test_result["comment"] == "Testing block<br>Step 2 blocked"
        assert test_result["scriptResults"][0]["status"] == "Pass"
        assert test_result["scriptResults"][1]["status"] == "Blocked"

    def test_T6(self, pytester: pytest.Pytester, adaptavist: Adaptavist):
        """Test skipping a step."""
        pytester.makepyfile(
            """
            import pytest

            def test_T6(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(True)
                with meta_block(2) as mb_2:
                    pytest.skip("Testing skip")
        """
        )
        report = pytester.inline_run("--adaptavist")
        test_run_key, test_name = get_test_values(report)
        test_result = adaptavist.get_test_result(test_run_key, test_name)
        assert test_result["status"] == "In Progress"
        assert test_result["scriptResults"][0]["status"] == "Pass"
        assert test_result["scriptResults"][1]["status"] == "Not Executed"

    def test_T7(self, pytester: pytest.Pytester, adaptavist: Adaptavist):
        """Test comments."""
        pytester.makepyfile(
            """
            def test_T7(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(True, message_on_pass="testing pass comment")
                with meta_block(2) as mb_2:
                    mb_2.check(False, message_on_fail="testing fail comment")
                with meta_block(3) as mb_3:
                    raise ValueError("testing exception reporting")
        """
        )
        report = pytester.inline_run("--adaptavist")
        test_run_key, test_name = get_test_values(report)
        test_result = adaptavist.get_test_result(test_run_key, test_name)
        assert test_result["status"] == "Fail"
        assert test_result["scriptResults"][0]["status"] == "Pass"
        assert "testing pass comment" in test_result["scriptResults"][0]["comment"]
        assert test_result["scriptResults"][1]["status"] == "Fail"
        assert "testing fail comment" in test_result["scriptResults"][1]["comment"]
        assert test_result["scriptResults"][2]["status"] == "Fail"
        assert "testing exception reporting" in test_result["scriptResults"][2]["comment"]

    def test_T8(self, pytester: pytest.Pytester, adaptavist: Adaptavist):
        """Test attachments."""
        pytester.maketxtfile(first_file="foo")
        pytester.maketxtfile(second_file="bar")
        pytester.makepyfile(
            """
            def test_T8(meta_block):
                with open("first_file.txt", "rb") as f:
                    with meta_block() as mb:
                        f.seek(1)
                        mb.check(True, attachment=f)
                        mb.check(True, attachment="second_file.txt")
                with meta_block(1) as mb_1:
                    with open("first_file.txt", "rb") as f:
                        mb_1.check(True, attachment=f)
                        mb_1.check(True, attachment="second_file.txt")
        """
        )
        report = pytester.inline_run("--adaptavist")
        test_run_key, test_name = get_test_values(report)
        test_result = adaptavist.get_test_result_attachment(test_run_key, test_name)
        assert len(test_result) == 4
        assert test_result[0]["filename"] == "second_file.txt"
        assert test_result[0]["filesize"] == 3
        assert test_result[1]["filename"] == "first_file.txt"
        assert test_result[1]["filesize"] == 2
        assert test_result[2]["filename"] == "second_file.txt"
        assert test_result[2]["filesize"] == 3
        assert test_result[3]["filename"] == "first_file.txt"
        assert test_result[3]["filesize"] == 3

    def test_T9(self, pytester: pytest.Pytester, adaptavist: Adaptavist):
        """
        Test meta_block.Action.FAIL_CONTEXT.
        Expect that step 1 fails, attachment on step 1 and step 2 passes.
        """
        pytester.maketxtfile(first_file="foo")
        pytester.maketxtfile(second_file="bar")
        pytester.makepyfile(
            """
            def test_T9(meta_block):
                with meta_block() as mb:
                    mb.check(True)
                with meta_block(1) as mb_1:
                    mb_1.check(False)
                    mb_1.check(True, attachment="second_file.txt")
                with meta_block(2) as mb_2:
                    mb_2.check(True)
        """
        )
        report = pytester.inline_run("--adaptavist")
        test_run_key, test_name = get_test_values(report)
        attachments = adaptavist.get_test_result_attachment(test_run_key, test_name)
        test_result = adaptavist.get_test_result(test_run_key, test_name)
        assert len(attachments) == 1
        assert test_result["status"] == "Fail"
        assert test_result["scriptResults"][0]["status"] == "Fail"
        assert test_result["scriptResults"][1]["status"] == "Pass"

    def test_T10(self, pytester: pytest.Pytester, adaptavist: Adaptavist):
        """
        Test meta_block.Action.STOP_CONTEXT.
        Expect that step 1 is blocked, step 2 passes and overall test result is blocked.
        """
        pytester.maketxtfile(first_file="foo")
        pytester.maketxtfile(second_file="bar")
        pytester.makepyfile(
            """
            def test_T10(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(False, action_on_fail=mb_1.Action.STOP_CONTEXT)
                    mb_1.check(True, attachment="second_file.txt")
                with meta_block(2) as mb_2:
                    mb_2.check(True)
        """
        )
        report = pytester.inline_run("--adaptavist")
        test_run_key, test_name = get_test_values(report)
        attachments = adaptavist.get_test_result_attachment(test_run_key, test_name)
        test_result = adaptavist.get_test_result(test_run_key, test_name)
        assert len(attachments) == 0
        assert test_result["status"] == "Blocked"
        assert test_result["comment"] == "Step 1 blocked"
        assert test_result["scriptResults"][0]["status"] == "Blocked"
        assert test_result["scriptResults"][1]["status"] == "Pass"

    def test_T11(self, pytester: pytest.Pytester, adaptavist: Adaptavist):
        """
        Test meta_block.Action.FAIL_METHOD.
        Expect that step 1 is failed, no attachment at step 1, step 2 not executed and overall test result is failed.
        """
        pytester.maketxtfile(first_file="foo")
        pytester.maketxtfile(second_file="bar")
        pytester.makepyfile(
            """
            def test_T11(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(False, action_on_fail=mb_1.Action.FAIL_METHOD)
                    mb_1.check(True, attachment="second_file.txt")
                with meta_block(2) as mb_2:
                    mb_2.check(True)
        """
        )
        report = pytester.inline_run("--adaptavist")
        test_run_key, test_name = get_test_values(report)
        attachments = adaptavist.get_test_result_attachment(test_run_key, test_name)
        test_result = adaptavist.get_test_result(test_run_key, test_name)
        assert len(attachments) == 0
        assert test_result["status"] == "Fail"
        assert test_result["comment"] == "Step 1 failed:"
        assert test_result["scriptResults"][0]["status"] == "Fail"
        assert test_result["scriptResults"][1]["status"] == "Not Executed"

    def test_T12(self, pytester: pytest.Pytester, adaptavist: Adaptavist):
        """
        Test meta_block.Action.STOP_METHOD.
        Expect that step 1 is blocked, no attachment at step 1, step 2 not executed and overall test result is blocked.
        """
        pytester.maketxtfile(first_file="foo")
        pytester.maketxtfile(second_file="bar")
        pytester.makepyfile(
            """
            def test_T12(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(False, action_on_fail=mb_1.Action.STOP_METHOD)
                    mb_1.check(True, attachment="second_file.txt")
                with meta_block(2) as mb_2:
                    mb_2.check(True)
        """
        )
        report = pytester.inline_run("--adaptavist")
        test_run_key, test_name = get_test_values(report)
        attachments = adaptavist.get_test_result_attachment(test_run_key, test_name)
        test_result = adaptavist.get_test_result(test_run_key, test_name)
        assert len(attachments) == 0
        assert test_result["status"] == "Blocked"
        assert test_result["scriptResults"][0]["status"] == "Blocked"
        assert test_result["scriptResults"][1]["status"] == "Not Executed"

    def test_T13(self, pytester: pytest.Pytester, adaptavist: Adaptavist, test_run: str):
        """
        Test meta_block.Action.FAIL_SESSION.
        Expect that T13 is failed, no attachment at T13, T12 is set to blocked. T11 is untouched and status In Progress
        """
        pytester.maketxtfile(first_file="foo")
        pytester.maketxtfile(second_file="bar")
        pytester.makepyfile(
            """
            def test_T11(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(True)
            def test_T13(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(False, action_on_fail=mb_1.Action.FAIL_SESSION)
                    mb_1.check(True, attachment="second_file.txt")
                with meta_block(2) as mb_2:
                    mb_2.check(True)
            def test_T12(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(True)
        """
        )
        report = pytester.inline_run("--adaptavist")
        _, test_name = get_test_values(report, "test_T13")
        test_result = adaptavist.get_test_result(test_run, test_name)
        assert test_result["status"] == "Fail"
        _, test_name = get_test_values(report, "test_T12")
        test_result = adaptavist.get_test_result(test_run, test_name)
        assert test_result["status"] == "Blocked"
        _, test_name = get_test_values(report, "test_T11")
        test_result = adaptavist.get_test_result(test_run, test_name)
        assert test_result["status"] == "In Progress"  # Ensure that T11 is untouched

    def test_T14(self, pytester: pytest.Pytester, adaptavist: Adaptavist, test_run: str):
        """
        Test meta_block.Action.STOP_SESSION.
        Expect that T14 is blocked, step 2 not executed. T12 also set to blocked
        """
        pytester.maketxtfile(first_file="foo")
        pytester.maketxtfile(second_file="bar")
        pytester.makepyfile(
            """
            def test_T14(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(False, action_on_fail=mb_1.Action.STOP_SESSION)
                with meta_block(2) as mb_2:
                    mb_2.check(True)
            def test_T12(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(True)
        """
        )
        report = pytester.inline_run("--adaptavist")
        _, test_name = get_test_values(report, "test_T14")
        test_result = adaptavist.get_test_result(test_run, test_name)
        assert test_result["status"] == "Blocked"
        _, test_name = get_test_values(report, "test_T12")
        assert test_result["status"] == "Blocked"

    def test_T15(self, pytester: pytest.Pytester, adaptavist: Adaptavist, test_run: str):
        """
        Test meta_block.Action.STOP_EXIT_SESSION.
        Expect that T15 is blocked. T12 not in test result --> not in test cycle in ATM
        """
        pytester.maketxtfile(first_file="foo")
        pytester.maketxtfile(second_file="bar")
        pytester.makepyfile(
            """
            def test_T15(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(False, action_on_fail=mb_1.Action.STOP_EXIT_SESSION)
                with meta_block(2) as mb_2:
                    mb_2.check(True)
            def test_T12(meta_block):
                with meta_block(1) as mb_1:
                    mb_1.check(True)
        """
        )
        pytester.inline_run("--adaptavist")
        test_name = test_run.split("-")[0] + "-T15"
        test_result = adaptavist.get_test_result(test_run, test_name)
        assert test_result["status"] == "Blocked"
        test_name = test_run.split("-")[0] + "-T12"
        test_result = adaptavist.get_test_result(test_run, test_name)
        assert test_result == {}

    def test_T16(self, pytester: pytest.Pytester, adaptavist: Adaptavist, test_run: str):
        """
        Test meta_data
        Expect that T16 is failed. Upload file with correct filename and size.
        """
        pytester.maketxtfile(first_file="foo")
        pytester.maketxtfile(second_file="bar")
        pytester.makepyfile(
            """
            import io
            def test_T16(meta_data):
                meta_data["comment"] = "unexpected result"
                attachment = io.StringIO()
                attachment.write("this is just a simple attachment")
                meta_data["attachment"] = attachment
                meta_data["filename"] = "content.txt"
                assert False
        """
        )
        pytester.inline_run("--adaptavist")
        test_name = test_run.split("-")[0] + "-T16"
        test_result = adaptavist.get_test_result(test_run, test_name)
        assert test_result["status"] == "Fail"
        attachments = adaptavist.get_test_result_attachment(test_run, test_name)
        assert attachments[0]["filename"] == "content.txt"
        assert attachments[0]["filesize"] == 32



"""Constants used by the module."""

COLORMAP = {
    "passed": {"green": True, "bold": True},
    "failed": {"red": True, "bold": True},
    "blocked": {"blue": True, "bold": True},
    "skipped": {"yellow": True, "bold": True},
}

META_BLOCK_TIMEOUT = 600

TEST_CYCLE_NAME_DEFAULT = "%(project_key) %(test_run_suffix)"
TEST_PLAN_NAME_DEFAULT = "%(project_key) %(test_plan_suffix)"
